#include <pthread.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "structs.h"
#include "core.h"
#include "threads.h"
#include "ocsp_codes.h"
#include "ocsp_request.h"
#include "ocsp_response.h"

extern OCSPD_CONFIG *ocspd_conf;

extern pthread_mutex_t clifd_mutex;
extern pthread_cond_t clifd_cond;

extern pthread_mutex_t crl_mutex;

int thread_make ( int i ) {
	Thread *th_ptr = NULL;
	int ret;

	th_ptr = &(ocspd_conf->threads_list[i]);
	if( (ret=pthread_create( &ocspd_conf->threads_list[i].thread_tid, 
			NULL, thread_main, (void *) i )) ) {

		syslog(LOG_ERR,"ERROR::OPENCA_SRV_ERR_THREAD_CREATE");
		return(ret);
	}

	return(OCSPD_SRV_OK);
}

void * thread_main ( void *arg ) {
	int connfd;

	OCSP_REQUEST  * req = NULL;
	OCSP_RESPONSE * resp = NULL;

	struct sigaction sa;

	if ( ocspd_conf->verbose ) {
		syslog(LOG_INFO, "INFO::OPENCA_SRV_INFO_TREAD::%s",
			"new thread created");
	}

	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;

	if (sigaction(SIGPIPE, &sa, NULL) == -1) {
		perror("Error during handling the death processes");
		exit(1);
	}

	for ( ; ; ) {
		pthread_mutex_lock( &clifd_mutex );
		while ( ocspd_conf->iget == ocspd_conf->iput ) {
			pthread_cond_wait( &clifd_cond, &clifd_mutex );
			if( ocspd_conf->debug ) {
				fprintf(stderr, 
					"[Thread] - awaken (iget=%d, iput=%d)\n",
					ocspd_conf->iget, ocspd_conf->iput );
			}
		}
		connfd = ocspd_conf->clifd[ocspd_conf->iget];
		if( ocspd_conf->debug )  {
			fprintf(stderr, "[Thread] - got connfd (%d)\n", connfd);
		}
		/*
		fprintf( stderr, "TREAD::%d got connfd=%d from slot %d!\n",
				(int) arg, connfd, ocspd_conf->iget);
		*/

		ocspd_conf->iget += 1;
		if( ocspd_conf->iget == ocspd_conf->nthreads ) 
			ocspd_conf->iget = 0;

		pthread_mutex_unlock( &clifd_mutex );
		// fprintf(stderr, "[Thread] - released clifd_mutex\n");
		ocspd_conf->threads_list[ (int) arg ].thread_count++;

		if((req = ocspd_req_get_socket( connfd, ocspd_conf )) == NULL ) {
			syslog(LOG_ERR, "ERROR::Can not parse REQ");
			goto err;
		}

		if( ocspd_conf->debug ) {
			fprintf(stderr, "[Thread] - read response from socket\n");
		}

		if(( make_ocsp_response(&resp, req, ocspd_conf)) == 0 ) {
                        syslog( LOG_ERR, "ERROR::Can not generate response!" );
                        goto err;
		}

		if( ocspd_conf->debug) {
			fprintf(stderr, "[Thread] - built response from socket\n");
		}

		goto end;

err:
		if( ocspd_conf->debug) {
			fprintf(stderr, "[Thread] - An error occurred!\n");
		}

		if((resp = OCSP_response_create(
				OCSP_RESPONSE_STATUS_MALFORMEDREQUEST, NULL)) == NULL) {
			close( connfd );
			continue;
		}

end:
                ocspd_resp_send_socket( connfd, resp );
		// fprintf(stderr, "[Thread] - sent resp on socket\n");

		if( resp ) OCSP_RESPONSE_free (resp);
		if( req ) OCSP_REQUEST_free (req);

		close( connfd );
		// fprintf(stderr, "[Thread] - closed connfd socket\n");
	}
}
